using System.Reflection;
using GraphicsManager;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects;
using Luski.GUI.MainScreen.UI.LuskiControls;
using Luski.net.Structures.Public;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace Luski.GUI.MainScreen.UI.PublicServers;

public class PublicChat : UserControl
{

    public FlowLayout MessageFlow;
    private Label title, desc;
    private TextBox tb;
    private SocketChannel? Channel;
    UserControl titlecon;
    private Rectangle? UserCon;
    private SocketMessage? first;
    private FlowLayout? memberflow = null;
    
    public PublicChat()
    {
        
        base.Size = new(980.ScaleInt(), 866.ScaleInt());
        BackgroundColor = new(40,40,40,255);
        Anchor = ObjectAnchor.All;
        Controls.Add(MessageFlow = new()
        {
            Size = new(base.Size.X, 761.ScaleInt()),
            Location = new(0, 52.ScaleInt(), 0),
            BackgroundColor = new(40,40,40,255),
            Anchor = ObjectAnchor.All,
            HScrollPixels = Globals.Settings.PerScrollPixels,
        });
        MessageFlow.FlowUpdate += MessageFlowOnFlowUpdate;
        
        Controls.Add(titlecon = new UserControl()
        {
            Anchor = ObjectAnchor.Left | ObjectAnchor.Top | ObjectAnchor.Right,
            Size = new(980.ScaleInt(), 48.ScaleInt()),
            BackgroundColor = new(50,50,50,255),
        });
        if (LuskiExperiments.ServerExperiments.MemberList.IsEnabled())
        {
            UserCon =
                new(Globals.ms.TextureManager.GetTextureResource("person.png"))
                {
                    Size = new(24.ScaleInt()),
                    Location = new(944.ScaleInt(), 12.ScaleInt(),0),
                    Anchor = ObjectAnchor.Right | ObjectAnchor.Top,
                    Shader = Rectangle.DefaultAlphaShader[Globals.ms.Context],
                    BackgroundColor = Color4.Red
                };
            UserCon.Clicked += UserConOnClicked;
            titlecon.Controls.Add(UserCon);
        }
        LuskiExperiments.ServerExperiments.MemberList.EventToggled += MemberListOnEventToggled;
        
        titlecon.ForceDistanceUpdate(this);
        titlecon.Controls.Add(title = new Label(Globals.DefaultFont)
        {
            //Location = new(
        });
        titlecon.Controls.Add(desc = new Label(Globals.DefaultFont)
        {
            Color = new(161,161,161,255), 
            Location = new(title.Location.X + title.Size.X + 5, title.Location.Y, 0)
        });
        
        Controls.Add(tb = new()
        {
           // InsideColor = new(28, 28, 28, 255),
            //BorderColor = Color4.DarkCyan,
            Location = new(10.ScaleInt(), 824.ScaleInt(), 0),
            Size = new(960.ScaleInt(), 34.ScaleInt()),
            Anchor = ObjectAnchor.Bottom | ObjectAnchor.Left | ObjectAnchor.Right,
            HoverMouse = MouseCursor.IBeam,
            //Shader = Rectangle.DefaultAlphaShader[Globals.ms.Context],
            //Shader = Rectangle.DefaultAlphaShader[Globals.ms.Context],
            BackgroundColor = Color4.Red,
            TextLocation = TextLocation.PostiveTureCenterLeft
        });
        tb.KeyPress += TbOnKeyPress;
        tb.OnRemoveLine += TbOnOnRemoveLine;
        tb.OnNewLine += TbOnOnNewLine;
        tb.ForceDistanceUpdate(this);
        //tb.KeyPress += TbOnKeyPress;
        //Globals.Luski.MainServer.MessageReceived += LuskiOnMessageReceived;
    }

    private async Task MessageFlowOnFlowUpdate(bool arg1, uint arg2, uint arg3)
    {
        if (!loadingm && arg1 && arg3 == 0 && (arg2 != 0 || MessageFlow.MaxScrollValue == 0))
        {
            loadingm = true;
            var messages = await Channel!.GetMessages(CancellationToken.None, first!, Globals.Settings.LoadPerChannel);
            if (messages.Count == 0) return;
            index = 0;
            fakeIndex = false;
            lastmIndex = null;
            lastUserIndex = null;
            LastChatMessageIndex = null;
            await AddMessagesIndex(messages);
            MessageFlow.ForceScrollUpdate();
            first = messages.Last();
            if (Globals.Settings.LoadPerChannel == messages.Count)loadingm = false;
        }
    }

    private async Task TbOnKeyPress(KeyboardKeyEventArgs arg)
    {
        if (arg.Key == Keys.Enter && !arg.Shift)
        {
            await Channel!.SendMessage(tb.Text);
            tb.Text = string.Empty;
            tb.CursorLocation = 0;
        }
    }

    private Task TbOnOnRemoveLine()
    {
        BlockDraw = true;
        tb.Location = new(tb.Location.X, tb.Location.Y + (int)tb.Font.PixelHeight, 0);
        MessageFlow.Size = new(MessageFlow.Size.X, MessageFlow.Size.Y + (int)tb.Font.PixelHeight);
        BlockDraw = false;
        return Task.CompletedTask;
    }

    private Task TbOnOnNewLine()
    {
        BlockDraw = true;
        tb.Location = new(tb.Location.X, tb.Location.Y - (int)tb.Font.PixelHeight, 0);
        MessageFlow.Size = new(MessageFlow.Size.X, MessageFlow.Size.Y - (int)tb.Font.PixelHeight);
        BlockDraw = false;
        return Task.CompletedTask;
    }

    private async Task MemberListOnEventToggled(bool arg)
    {
        if (arg)
        {
            UserCon =
                new(Globals.ms.TextureManager.GetTextureResource("person.png"))
                {
                    Size = new(24.ScaleInt()),
                    Location = new(base.Size.X - 36.ScaleInt(), 12.ScaleInt(),0),
                    Anchor = ObjectAnchor.Right | ObjectAnchor.Top,
                    Shader = Rectangle.DefaultAlphaShader[Globals.ms.Context],
                    BackgroundColor = Color4.Red
                };
            UserCon.Clicked += UserConOnClicked;
            titlecon.Controls.Add(UserCon);
        }
        else
        {
            if (um_open) await UserConOnClicked(UserCon!);
            titlecon.Controls.Remove(UserCon!);
            UserCon = null;
        }
    }

    private bool um_open = false;

    private async Task UserConOnClicked(IRenderObject arg)
    {
        um_open = !um_open;
        if (um_open)
        {
            if (memberflow is null)
            {
                int x = 232.ScaleInt();
                memberflow = new()
                {
                    BackgroundColor = new(255, 0, 0, 255),
                    Size = new(x, Size.Y - titlecon.Size.Y),
                    Location = new(Size.X - x, titlecon.Size.Y, 0)
                };
                Controls.Add(memberflow);
                Dictionary<long, RoleView> Roles = new();
            
                foreach (var m in await Channel!.GetMembers())
                {
                    Role top_role = (await m.GetRoles())[0];
                    if (!Roles.ContainsKey(top_role.ID))
                    {
                        Roles.Add(top_role.ID, new(top_role));
                        memberflow.Controls.Add(Roles[top_role.ID]);
                    }
                    Roles[top_role.ID].AddUser(m);
                }
            }

            memberflow.Visible = true;

            MessageFlow.Size = new(MessageFlow.Size.X - memberflow.Size.X, MessageFlow.Size.Y);
            tb.Size = new(tb.Size.X - memberflow.Size.X, tb.Size.Y);
            tb.ForceDistanceUpdate();
        }
        else
        {
            memberflow!.Visible = false;
            MessageFlow.Size = new(MessageFlow.Size.X + memberflow.Size.X, MessageFlow.Size.Y);
            tb.Size = new(tb.Size.X + memberflow.Size.X, tb.Size.Y);
            tb.ForceDistanceUpdate();
        }
    }

    private SocketMessage? lastm;
    private long? lastUser;
    private bool fake = false;
    public ChatMessage? LastChatMessage;
    
    private SocketMessage? lastmIndex;
    private long? lastUserIndex;
    private bool fakeIndex = false;
    public ChatMessage? LastChatMessageIndex;
    private bool loadingm = false;

    public async Task LoadChannel(SocketChannel channel)
    {
        Channel = channel;
        if (memberflow is not null)
        {
            memberflow.Controls.Clear();
        }
        title.Text = channel.Name;
        desc.Text = channel.Description;
        desc.Location = new((int)(title.Location.X + title.Size.X + 5.ScaleInt()), desc.Location.Y, 0);
        if (Window is not null)
        {
            Window.Title = $"{channel.Name} | {channel.Server.Name} - Luski";
        }
        tb.WatermarkText = $"Message {channel.Name}";
        tb.Text = "";
        tb.CursorLocation = 0;
        tb.Focus();
    }

    public void ClearChat()
    {
        MessageFlow.Controls.Clear();
        MessageFlow.ScrollUpdatesInterval = 33;
        lastm = null;
        lastUser = null;
        LastChatMessage = null;
        first = null;
    }

    public async Task AddMessages(IEnumerable<SocketMessage> messages, bool reverse = true)
    {
        MessageFlow.BlockDraw = true;
        if (reverse)
        {
            foreach (SocketMessage message in messages.Reverse())
            {
                await AddMessage(message);
            }
        }
        else
        {
            foreach (SocketMessage message in messages)
            {
                await AddMessage(message);
            }
        }
        MessageFlow.BlockDraw = false;
    }
    
    public async Task AddMessagesIndex(IEnumerable<SocketMessage> messages)
    {
        MessageFlow.BlockDraw = true;
        index = 0;
        foreach (SocketMessage message in messages.Reverse())
        {
            await AddMessageIndex(message);
        }
        MessageFlow.BlockDraw = false;
    }
    
    public async Task AddMessage(SocketMessage Message, bool scrool = false)
    {
        if (first is null) first = Message;
        bool hasbeentenmin = false;
        if (lastm is not null)
            hasbeentenmin =
                Channel!.Epoch.AddMilliseconds(lastm.ID >> 22).ToLocalTime().AddMinutes(10) <
                Channel!.Epoch.AddMilliseconds(Message.ID >> 22).ToLocalTime();
        lastm = Message;
        if (lastUser is null || lastUser != Message.AuthorID || hasbeentenmin || fake != Message.IsProfile)
        {
            
            if (Window is null || !Globals.ms.InvokeRequired)
            {
                MessageFlow.Controls.Add(LastChatMessage = await ChatMessage.MakeChatMessage(this, Message));
                if (scrool) MessageFlow.ScrollToBottom();
            }
            else
            {
                LastChatMessage = await ChatMessage.MakeChatMessage(this, Message);
                Globals.ms.Invoke(() =>
                {
                    MessageFlow.Controls.Add(LastChatMessage);
                    if (scrool) MessageFlow.ScrollToBottom();
                    Window.TryDraw();
                });
            }
        }
        else
        {
            if (Window is null || !Globals.ms.InvokeRequired)
            {
                await LastChatMessage!.AddMessage(Message);
                if (scrool) MessageFlow.ScrollToBottom();
            }
            else
            {
                Globals.ms.Invoke(() =>
                {
                    LastChatMessage!.AddMessage(Message);
                    if (scrool) MessageFlow.ScrollToBottom();
                    Window!.TryDraw();
                });
            }
        }
        fake = Message.IsProfile;
        lastUser = Message.AuthorID;
    }

    private int index = 0;
    
    public async Task AddMessageIndex(SocketMessage Message)
    {
        bool hasbeentenmin = false;
        if (lastmIndex is not null)
            hasbeentenmin =
                Channel!.Epoch.AddMilliseconds(lastmIndex.ID >> 22).ToLocalTime().AddMinutes(10) <
                Channel!.Epoch.AddMilliseconds(Message.ID >> 22).ToLocalTime();
        lastmIndex = Message;
        if (lastUserIndex is null || lastUserIndex != Message.AuthorID || hasbeentenmin || fake != Message.IsProfile)
        {
            if (Window is null || !Globals.ms.InvokeRequired)
            {
                MessageFlow.Controls.Insert(index, LastChatMessageIndex = await ChatMessage.MakeChatMessage(this, Message));
                LastChatMessageIndex.LoadToParent(MessageFlow, Globals.ms);
                index++;
            }
            else
            {
                LastChatMessageIndex = await ChatMessage.MakeChatMessage(this, Message);
                Globals.ms.Invoke(() =>
                {
                    MessageFlow.Controls.Insert(index, LastChatMessageIndex);
                    LastChatMessageIndex.LoadToParent(MessageFlow, Globals.ms);
                    index++;
                    Window.TryDraw();
                });
            }
        }
        else
        {
            if (Window is null || !Globals.ms.InvokeRequired)
            {
                await LastChatMessageIndex!.AddMessage(Message);
            }
            else
            {
                Globals.ms.Invoke(() =>
                {
                    LastChatMessageIndex!.AddMessage(Message);
                    Window!.TryDraw();
                });
            }
        }
        fakeIndex = Message.IsProfile;
        lastUserIndex = Message.AuthorID;
    }
}